
/*
    Provides a PETSc interface to SUNDIALS. Alan Hindmarsh's parallel ODE
   solver developed at LLNL.
*/

#if !defined(__PETSCSUNDIALS_H)
#define __PETSCSUNDIALS_H

#include <petsc-private/tsimpl.h>       /*I   "petscts.h"   I*/
#include <petsc-private/pcimpl.h>
#include <petsc-private/matimpl.h>

/*
   Include files specific for SUNDIALS
*/
#if defined(PETSC_HAVE_SUNDIALS)

EXTERN_C_BEGIN
#include "mpi.h"
#include "cvode/cvode.h"                  /* prototypes for CVODE fcts. */
#include "cvode/cvode_impl.h"
#include "cvode/cvode_spgmr.h"            /* prototypes and constants for CVSPGMR solver */
#include "nvector/nvector_parallel.h"     /* definition N_Vector and macro NV_DATA_P  */
EXTERN_C_END

#define CV_SS  1
#define CV_SV  2
#define CV_WF  3
#define CV_NULL_Y0 -39
#define	CV_BAD_ITOL -40
#define	CV_NULL_F -41
#define	CV_BAD_NVECTOR -42
#define	CV_NULL_ABSTOL -43
#define	CV_BAD_RELTOL -44
#define	CV_BAD_ABSTOL -45
#define ZERO RCONST(0.0)
#define ONE RCONST(1.0)
#define ETAMX1 RCONST(10000.0)
#define	MSGCV_BAD_ITOL "Illegal value for itol. The legal values are CV_SS, CV_SV, and CV_WF."
#define lrw1 (cv_mem->cv_lrw1)
#define liw1 (cv_mem->cv_liw1)
#define ewt (cv_mem->cv_ewt)
#define acor (cv_mem->cv_acor)
#define tempv (cv_mem->cv_tempv)
#define ftemp (cv_mem->cv_ftemp)
#define qmax (cv_mem->cv_qmax)
#define zn (cv_mem->cv_zn)
#define lrw (cv_mem->cv_lrw)
#define liw (cv_mem->cv_liw)

static booleantype CVCheckNvector(N_Vector tmpl);

static booleantype CVAllocVectors(CVodeMem cv_mem, N_Vector tmpl);

int CVodeSetFdata(void *cvode_mem, void *f_data);

int CVodeMalloc(void *cvode_mem, CVRhsFn f, realtype t0, N_Vector y0, 
                int itol, realtype reltol, void *abstol);

typedef struct {
  Vec update;           /* work vector where new solution is formed */
  Vec ydot;             /* work vector the time derivative is stored */
  Vec w1,w2;            /* work space vectors for function evaluation */

  /* PETSc peconditioner objects used by SUNDIALS */
  PetscInt                  cvode_type;   /* the SUNDIALS method, BDF or ADAMS  */
  TSSundialsGramSchmidtType gtype;
  PetscReal                 linear_tol;
  PetscReal                 mindt,maxdt;

  /* Variables used by Sundials */
  MPI_Comm  comm_sundials;
  double    reltol;
  double    abstol;          /* only for using SS flag in SUNDIALS */
  N_Vector  y;               /* current solution */
  void      *mem;
  PetscBool monitorstep;     /* flag for monitor internal steps; itask=V_ONE_STEP or itask=CV_NORMAL*/
  PetscInt  maxl;            /* max dimension of the Krylov subspace to be used */
} TS_Sundials;
#endif

#endif




